// Copyright (c) 2000 National Semiconductor Corporation, All Rights Reserved.

/* This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the License, or (at your option) any later version. 

   This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
   warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
   for more details. 

   You should have received a copy of the GNU General Public License along with this program; if not, write to the Free
   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
*/

#undef __NO_VERSION__

/******************************************************************************
*h*
*   Name:               dp.c
*
*   Description:        This is the source code for the Linux Network OS
*		                Specific Module
*
*   Author:     		NSC
*
*   History:

*h*
******************************************************************************/

#include "nsmtypes.h"
#include "dp83820.h"
#include "hsm.h"
#include "nsm.h"

struct device *DevList = NULL;

/******************************************************************************
*f*
*   Name:           DrvProbe
*
*   Description     This routine enumerates Drv ethernet devices on the PCI 
*   				bus, and registers each device found.
*
*   Parameters:     Pointer to the device structure 
*
*   Return Value:   OK or -ENODEV 
*
*f*
******************************************************************************/

int
DrvProbe (struct device *dev)
{
	int DevCount=0;
	struct pci_dev *pPciDev=NULL;
	UINT  iobase, membase;
	AdapterContext *pNsc;
	NsmContext *pNsm;

#ifdef NSCDEBUG
	UINT	cfgcs_val;
#endif

	/* Check for PCI BIOS being present, return ENODEV if not present */
	if (! pci_present())
	{
		printk("\nPCI bios not present"); 
	
		/* No such device */
   		return -ENODEV; 
	}

	/* Find the instances of the device present, and for each instance of the 
	device do the following */

	while ( (pPciDev = pci_find_device (PCI_VENDOR_ID_NS, 
		PCI_DEVICE_ID_83820, pPciDev) ) != NULL)
	{

#ifdef NSCDEBUG
		printk(KERN_INFO "\nFound the NSC device");
#endif
 
		/* Get the IOBASE, MEMBASE address  */
		iobase = (UINT) pPciDev->base_address[0];
		membase = (UINT)pPciDev->base_address[1];

		iobase &= PCI_BASE_ADDRESS_IO_MASK;
		membase &= PCI_BASE_ADDRESS_MEM_MASK;

#ifdef IOMAPPED_IO
		
		/* Check if a range of ports is already locked by other drivers */
		if (check_region(iobase,PCI_IO_SIZE))
		{  
			printk ("check_region failed : IO region already in use\n");
			continue;
		}
#endif		
		/* Allocate space for the device structure, name the device, and also 
		for the private data structure, in this case, the NsmContext. Add the 
		device to the ethernet device list and register the network 
		device. */

		dev = init_etherdev(dev, sizeof(NsmContext));

		if (dev == NULL) 
		{	
			printk (KERN_ERR "%s: Failed to create device struct\n",DRV_NAME);

			break;
		}

		printk(KERN_INFO "\nDP83820 Gigabit NIC : %s\n",dev->name);

		/* Get the pointer to the NsmContext */
		pNsm = (NsmContext *)dev->priv;
		
		pNsm->dev = dev;

		/* Get the pointer to the adapter context */
		pNsc = &pNsm->AdpCtxt ;
		
		/* Allocate the structure to hold the enet_statistics  */
		/* This is reqd here because an ifconfig done can request the statistcs 
			even before an open is done */
		pNsm->pAdapterStats = (struct enet_statistics *) 
			kmalloc(sizeof(struct enet_statistics), GFP_KERNEL);	

		/* Check if the allocation was successful */
		if(pNsm->pAdapterStats == NULL)
		{
			printk(KERN_INFO "Allocation of Statistics struct failed\n");
			unregister_netdev(dev);
			kfree(dev);
			break;
		}

		pNsc->pNsmContext = (void *)pNsm;

		/* initialize the device data like iobase, irq, and the entry points 
		for the driver in the device structure */

#ifdef NSCDEBUG
		pci_read_config_dword(pPciDev,PCI_COMMAND,&cfgcs_val);  
		printk("The CFGCS is %x",cfgcs_val); 
#endif

		pNsc->CacheLineSize = CACHE_LINE_SIZE;

		pNsc->MaxPktSize = BUF_SIZE_MAX;

#ifdef MEMMAPPED_IO

		dev->base_addr = membase; /* Memory mapped IO */
		pNsc->RegAddr = ioremap(membase, PCI_MEM_SIZE);

#elif IOMAPPED_IO

		dev->base_addr = iobase;
		pNsc->RegAddr =  dev->base_addr;
#endif

#ifdef NSCDEBUG
		printk(KERN_INFO "The IRQ value is %d\n",pPciDev->irq);
#endif
	
		dev->irq       = pPciDev->irq;
		dev->open      = DrvOpen;
    	dev->stop      = DrvClose;
    	dev->get_stats = DrvGetStatistics;
		dev->do_ioctl  = DrvIoctl;
		dev->hard_start_xmit = DrvSend;
		dev->set_multicast_list = DrvSetMulticastList;

		/* initialise the ethernet specific generic fields */
		ether_setup (dev); 

		/* Change the mtu function in the device structure as ether_setup sets 
		it to a function which can accepts MTUs of size less than 1500, so 
		change it to a function which can change the MTU to any value 
		less than 65280 and greater than 68. */
		
		dev->change_mtu = ChangeMtu; 

#ifdef IOMAPPED_IO
		
		/* reserve IO region, this locks the ports for later use */
		request_region (iobase, PCI_IO_SIZE, dev->name);

#endif

		/* Chain the device */
		pNsm->next = (void *)DevList;
   		DevList = dev;

#ifdef NSCDEBUG
		printk(KERN_INFO "\nDev structure ptr is : %p\n",dev);
#endif
		
		/* Update Counters for the number of devices found */
		DevCount++;
		
		dev=NULL;
		/* pPciDev = NULL;  */
  	}

	if(DevCount == 0)
	{
		printk (KERN_ERR "No device of NSC present\n");
		return -ENODEV;
	}
			
	return 0;

}

/******************************************************************************
*f*
*   Name:           DrvOpen
*
*   Description:    DrvOpen opens and initializes a device. This is 
*    				typically called when an ifconfig is done for this 
*					interface.
*
*   Parameters:     Pointer to the device structure 
*
*   Return Value:   OK or -EAGAIN 
*
*f*
******************************************************************************/

STATIC int
DrvOpen (struct device *dev)
{

	UINT stat;
	NsmContext *pNsm;
	AdapterContext *pNsc;
	UINT dp_val;
	
#ifdef FT_DBG
	static UINT CallCount=0;
	printk("The DrvOpen call No : %d",CallCount++);
#endif

	/* Get the pointer to the allocated private data structure, NsmContext, 
	from the device structure */
	pNsm = (NsmContext *)dev->priv;

	/* Get the pointer to the AdapterContext structure from the NsmContext */
	pNsc = &pNsm->AdpCtxt;

	/* Initialize fields which would indicate the capabilities of the hardware,
	 such as the, checksum offload */
	HsmInitContext(pNsc);

#ifdef NSCDEBUG
	printk(KERN_INFO "HsmInitContext called\n");
#endif

	/* Find out the capabilities of the OS and indicate whether the above 
	capabilties are supported by the OS. */
	
	pNsc->OsPrio = MAX_PRIORITY_QUEUES;
	pNsc->PhysCapabilities &= ~(VLAN_TAG_INSERTION_GEN_ON | 
		BIG_ENDIAN | TX_CHKSUM_OFFLOAD_ON_GEN | 
			SET_RX_CHKSUM_OFFLOAD_ON );
	
	/* Allocate the Tx and Rx lists, based on the number of priority levels 
	supported. Call the HsmInitialise routine of the HSM, this will also setup 
	transmit & receive control. This will set up descriptor lists for each 
	level of priorities on both send and receive side. */

	if ( (stat = HsmInitialize(pNsc) ) == FAILURE )
	{
		printk (KERN_ERR "%s: HsmInitialize Failed \n", DRV_NAME);
		return -EAGAIN;
	}

#ifdef NSCDEBUG
	
	printk("The value of the tail %p 's CMDSTS : %x" , pNsc->PriQue[0].pRcbListTail->pRxDp, pNsc->PriQue[0].pRcbListTail->pRxDp->CmdSts);

#endif


#ifdef NSCDEBUG
	printk("The number of RX priorities : %d",pNsc->RxNumPrio);
#endif

	/* Now set the Mac address in the device structure */
	NsmCopy((char *)pNsc->PermMacAddr, (char *)dev->dev_addr, 
		MAC_ADDR_LEN); 

	/* Install interrupt vector and register the interrupt service routine */
	if(request_irq (dev->irq, DrvIsr, SA_SHIRQ,
    	dev->name, dev) != OK)
	{
		printk (KERN_ERR "%s: request_irq Failed \n", DRV_NAME);
		HsmUnInitialize(pNsc);
		return -EAGAIN;
	}

	/* Reset the adapter  */
	if( ( stat =HsmReset(pNsc) ) == FAILURE)
	{
		printk (KERN_ERR "%s: Resetting the adapter, failed \n", 
			DRV_NAME);
		free_irq (dev->irq, dev);
		HsmUnInitialize(pNsc);
		return -EAGAIN;
	} 

	/* Call HsmOpen call to open the adapter */
	if( ( stat =HsmOpen(pNsc) ) == FAILURE)
	{
		printk (KERN_ERR "%s: Opening the adapter, failed \n", 
			DRV_NAME);
		free_irq (dev->irq, dev);
		HsmUnInitialize(pNsc);
		return -EAGAIN;
	}
	
	HsmRxFilter(pNsc, (ACCEPT_PERFECT_MATCH_ON | ACCEPT_ALL_BROADCAST_ON));

	/* Mark the start flag in the device structure */
	dev->start = DEV_STARTED;
	
	dev->tbusy = 0;
	dev->interrupt = 0;

	/* Set the transmit queue length in the device structure */
	dev->tx_queue_len = (TX_QUEUE_SZ*MAX_PRIORITY_QUEUES);

	NsmRegRead32(pNsm, (pNsc->RegAddr + RXDP), &dp_val);

#ifdef NSCDEBUG			
	NsmDbgMsg1("RXDP's initial Val : %x", dp_val);
#endif

	/* Increment module reference count, MOD_INC_USE_COUNT */
	MOD_INC_USE_COUNT;

    return OK;
}

/******************************************************************************
*f*
*   Name:           DrvClose
*
*   Description:    This closes the device and reclaims the resources
*
*   Parameters:     Pointer to the device structure 
*
*   Return Value:   OK or ERROR  
*
*f*
******************************************************************************/

STATIC int
DrvClose (struct device *dev)
{

	NsmContext *pNsm;
	AdapterContext *pNsc;

#ifdef FT_DBG
	static UINT CallCount=0;
	printk("The DrvClose call No : %d",CallCount++);
#endif

	dev->start = 0;
	dev->tbusy = 1;

	/* Get the pointer to the NsmContext and thereby to the AdapterContext, 
	from the device structure */
	pNsm = (NsmContext *)dev->priv;
	pNsc   = &(pNsm->AdpCtxt);

#ifdef NSCDEBUG
	printk(KERN_INFO "DrvClose called\n");
#endif

	/* Stop Tx & Rx by calling HsmClose */
	HsmClose(pNsc); 

	/* Uninstall the interrupt vector, and deregister the interrupt service 
	routine */
	free_irq (dev->irq, dev);

	/* Free the Tx and Rx lists by calling HsmUnInitialize */
	HsmUnInitialize(pNsc);  

	/* Decrement module reference count using MOD_DEC_USE_COUNT */
	MOD_DEC_USE_COUNT;

  	return OK;
}

/******************************************************************************
*f*
*   Name:           DrvSend
*
*   Description:    This sends a packet down to the HSM, to be sent on wire
*
*   Parameters:     Pointer to the device structure
*                   Pointer to the sk_buff which contains the data to be sent 
*                   down
*
*   Return Value:   OK or ERROR
*
*f*
******************************************************************************/


STATIC int
DrvSend (struct sk_buff *skb, struct device *dev)
{
	NsmContext 		*pNsm;
	EthHeader 		*pEthHeader;
	AdapterContext 	*pNsc;
	PktInfo			Pkt;
	UCHAR			status;
	UINT			retval=OK;

#ifdef FT_DBG
	UCHAR 			LoopBack=FALSE;
	UCHAR 			SendPackets = TRUE;
	struct sk_buff 	*SkbNew;
#endif

#ifdef MULTIPLE_PACKETS
	PktStruct		Pkts;
	UCHAR 			stat;
#endif

#ifdef NSCDEBUG
	printk("Entered Send routine\n");	
#endif


	/* Get the pointer to the NsmContext and thereby to the AdapterContext, 
	from the device structure */
	pNsm = (NsmContext *)dev->priv;
	pNsc = &(pNsm->AdpCtxt);

#ifdef FT_DBG
	printk("The length of the skbuff to be sent is : %d", skb->len);
#endif

	/* Test the tbusy flag in the device structure before sending packet down */
	if (test_and_set_bit(0, &dev->tbusy))
	{
		printk("The device is busy \n");
		return 1;
	}
	
	/* Set the transmit start time in the device structure */
	dev->trans_start = jiffies;

	/* Check for the loopback packets, if it's a loopback packet  */
	/* Get the ethernet header and then get the source and the destination 
	addresses and compare them to check whether it is a loopback packet */
	pEthHeader = (EthHeader *)skb->data;

#ifdef FT_DBG 
	/* Not reqd to check for broadcast, as the upper layer 
	does the loop back */
	/* Check if the packet is a multicast or a broadcast packet, if so */
	if( (stat = IsBrdCastOrMultiCast(dev, pEthHeader->DstAddr)) == SUCCESS)
	{
		LoopBack = TRUE;
	}

	/* If the source and the destination addresses are the same */
	else if(CMP_ETH_ADDR ( pEthHeader->DstAddr, pEthHeader->SrcAddr, 
		MAC_ADDR_LEN ))

	if(CMP_ETH_ADDR ( pEthHeader->DstAddr, pEthHeader->SrcAddr, 
		MAC_ADDR_LEN ))
	{
		LoopBack = TRUE;
		SendPackets = FALSE;
	}

	/* This has to be sent down and also looped back */
	if(LoopBack & TRUE) 
	{
		printk("Loopback packet \n");
		if(SendPackets & TRUE)
		{
			/* Clone the packet  */
			SkbNew = skb_clone(skb, GFP_KERNEL);

			/* Check for NULL */
			if(SkbNew == NULL)
			{
				dev_kfree_skb(skb);

				/* Update the statistics for the dropped packets */
				pNsc->MacStats.txDroppedCount++;
				
				clear_bit(0, &dev->tbusy);
				return ERROR;
			}

			/* Send the new sk_buff up  */

			netif_rx(SkbNew);

			/* Send the original sk_buff down*/  
		
		}

		/* Else just pass the sk_buff up without any cloning */
		else
		{
			netif_rx(skb);

			/* Update the statistics for the successfully sent packets */
			pNsc->MacStats.txOkCount++;
			
			clear_bit(0, &dev->tbusy);	
			return OK;
		}
	}

#endif
	/* Encapsulate the data which needs to be sent on the wire in a Packet 
	struct */

#ifdef SINGLE_PACKET

	CreatePacketInfo( skb, &Pkt);

#elif MULTIPLE_PACKETS

	CreatePacketStruct( skb, &Pkts);

#endif

	/* Get the priority of the packet, which is going to be sent down from the 
	priority field in the skbuff. */

	/* Call the HSM routine to send the packet on the wire with the priority 
	obtained from the priority field of the skbuff. */ 

#ifdef NSCDEBUG
	printk("The priority of the skb is : %d",skb->priority);
#endif

#ifdef SINGLE_PACKET
	
	status = HsmSendPacket(pNsc, &Pkt, 0);
	
	if(status == FAILURE)
	{
		printk("Could not send the packet \n");
		dev_kfree_skb((struct sk_buff *)(Pkt.pOSCtrlBlk));
		retval = ERROR;
	}

#elif MULTIPLE_PACKETS

	stat = HsmSendPkts ( pNsc, &Pkts, skb->priority ); 

	if(stat == 0)
	{
		printk("Could not send the packet \n");
		dev_kfree_skb((struct sk_buff *)(Pkt.pOSCtrlBlk));
		retval = ERROR;
	}

#endif
	
	clear_bit(0, &dev->tbusy);

#ifdef NSCDEBUG
	printk("Exiting Send routine\n");	
#endif

 	return retval;
}

/******************************************************************************
*f*
*   Name:           DrvSetMulticastList 
*
*   Description:    NsmSetMulticastList sets multicast filters & promiscuous 
*                   mode 
*
*   Parameters:     Pointer to the device structure
*
*   Return Value:   None 
*
*f*
******************************************************************************/


STATIC void
DrvSetMulticastList (struct device *dev)
{
	struct dev_mc_list 	*mc_list;
	MultiList 			*pMultiList;
	MultiList			*pMulti;
	MultiList 			*pMultiNext;
	UINT				FilterFlags=0;

	/* Get the pointer to the NsmContext and thereby to the AdapterContext, 
	from the device structure */
	AdapterContext *pNsc = &(((NsmContext *)dev->priv)->AdpCtxt);

#ifdef NSCDEBUG
	printk(KERN_INFO "Entered the muticast routine\n");
#endif
    
	/* Setup promiscuous mode */ 
	if (dev->flags & IFF_PROMISC)
	{

#ifdef NSCDEBUG
		printk(KERN_INFO "Setting the promiscuous mode\n");
#endif

		FilterFlags |= PROMISCUOUS_ON;
	}
 
	/* Receiving all multicast packets  */
	if (dev->flags & IFF_ALLMULTI) 
	{

#ifdef NSCDEBUG
		printk(KERN_INFO "Setting the ALL MULTI mode\n");
#endif

		FilterFlags |= ACCEPT_ALL_MULTICAST_ON;
	}		

	/* Call the HSM function to set the receive filters */
	if(FilterFlags)
		HsmRxFilter(pNsc, FilterFlags);

	/* Set the list of multicast addresses in the mc_list */
	mc_list = dev->mc_list;
	pMultiList = NULL;

	/* Get the multicast address list from the mc_list and allocate a 
	MultiList structure and put the multicast addresses into the MultiList 
	structure  */

	while(mc_list)
	{
		pMulti = (MultiList *)kmalloc(sizeof(MultiList), GFP_KERNEL);
		if(pMulti == NULL)
		{
			if(pMultiList)
			{
				pMulti = pMultiList;
				while(pMulti)
				{
					pMultiNext = pMulti->Next;
					kfree(pMulti);
					pMulti = pMultiNext;
				}
			}
			return;
		}
		NsmCopy(mc_list->dmi_addr, 
			pMulti->MulticastAddr,MAX_ADDR_LEN); 
		pMulti->Next = pMultiList;
		pMultiList = pMulti;
		mc_list = mc_list->next;		        	
	}

	/* Call the HSM function to add the list of multicast addresses to the 
	filter */
	HsmMulticastAdd(pNsc, pMultiList, TRUE);	

	/* Free up the list of MultiList structures */
	pMulti = pMultiList;
	while(pMulti)
	{
		pMultiNext = pMulti->Next;
		kfree(pMulti);
		pMulti = pMultiNext;
	}

#ifdef NSCDEBUG
	printk(KERN_INFO "Exiting the muticast routine \n");
#endif

}

/******************************************************************************
*f*
*   Name:           DrvIsr 
*
*   Description:    This is the interrupt handler of the driver
*
*   Parameters:    	irq: This represents the IRQ value for which this handler 
*                   was registered
*
*                  	dev_id:  This is a pointer to the device structure
*
*                  	regs:    This is a pointer to a snapshot of the processor's 
*                   context, before the processor entered the interrupt code
* 
*   Return Value:   None
*
*f*
***************************************************************************/

STATIC void
DrvIsr (int irq, void *dev_id, struct pt_regs *regs)
{
	/* Get the pointer to the device structure from the dev_id */
	struct device * dev = ( struct device *) dev_id;

	/* Get the pointer to the AdapterContext */
	AdapterContext *pNsc = &(((struct NsmInfo *)dev->priv)->AdpCtxt);

	/* Set the interrupt flag in the device structure  */
	dev->interrupt = DEV_INTERRUPTED;

	/* Call the HSMs interrupt service routine */
	HsmIsr (pNsc);

	/* Clear the interrupt flag in the device structure */
	dev->interrupt = DEV_SERVICED_INTERRUPT; 
}


/******************************************************************************
*f*
*   Name:           DrvIoctl 
*
*   Description:    This is the ioctl handler of the driver 
*
*   Parameters:     dev: This is a pointer to the device structure
*
*                   rq:  pointer to a copy of the structure passed by the user 
*                   for this command
*
*                   cmd: This indicates the ioctl command
*
*   Return Value:   OK or -EOPNOTSUPP 
*
*f*
***************************************************************************/

int 
DrvIoctl( struct device *dev, struct ifreq *rq, int cmd )
{

	/* Get the data for the ioctl from the ifreq structure */ 
	void *data = &(rq->ifr_data);

	/* Get the pointer to the AdapterContext */
	AdapterContext *pNsc = &((NsmContext *)(dev->priv))->AdpCtxt;

#ifdef NSCDEBUG
	printk(KERN_INFO "Ioctl routine entered \n");	
#endif

	/* Depending on the cmd perform the following */
	switch( cmd )
	{

		/* For setting interface address(dev->dev_addr) */
		case SIOCSIFADDR:

			/* Set the MAC address in the device structure and also the 
			AdapterContext, with the one obtained from the data */

			NsmCopy((const void *)data, (void *)dev->dev_addr, MAC_ADDR_LEN);
			NsmCopy((const void *)data, (void *)pNsc->CurrMacAddr, 
				MAC_ADDR_LEN);
			HsmSetMacAddress(pNsc, pNsc->CurrMacAddr); 
			break;	

		/* The ioctl implementation for sockets, defines 16 ioctls cmds as 
		private to the interface, SIOCSDEVPRIVATE, SIOCSDEVPRIVATE+ 1,.. 
		& so on  */

		case SIOCDEVPRIVATE: 
			break;
		default:
			return -EOPNOTSUPP;

	}
	return OK;
}

/******************************************************************************
*f*
*   Name:           DrvGetStatistics 
*
*   Description:    This is the routine used to get the statistical information
*                   regarding the transmission and reception of packets.
*
*   Parameters:     dev: This is a pointer to the device structure
*
*   Return Value:   Pointer to the statistics structure 
*
*f*
***************************************************************************/

struct enet_statistics *
DrvGetStatistics(struct device *dev)
{
	/* Get the pointer to the NsmContext and thereby to the AdapterContext, 
		from the device structure */
	NsmContext *pNsm = (NsmContext *)dev->priv;
	AdapterContext *pNsc = &(pNsm->AdpCtxt);

	struct enet_statistics *pAdpStats = pNsm->pAdapterStats; 

#ifdef NSCDEBUG
	printk(KERN_INFO "The Get stats routine entered\n");
#endif

	/* Update the fields in the StatInfo structure (MacStats) in the 
		AdapterContext, by reading the current values from the MIB 
		registers of the adapter */
	if(HsmGetStatistics(pNsc) == FAILURE) 
	{
		return NULL;
	}
	
	/* Get the statistics from the adapter context structure & set the values 
		of the statistics in the enet_statistics structure */

	pAdpStats->rx_packets = pNsc->MacStats.rxOkCount;
	pAdpStats->rx_errors = pNsc->MacStats.rxErrorCount;
	pAdpStats->rx_dropped = pNsc->MacStats.rxDroppedCount;
	pAdpStats->rx_crc_errors = pNsc->MacStats.rxCrcErrorCount;
	pAdpStats->rx_frame_errors = pNsc->MacStats.rxFaeCount;
	pAdpStats->rx_missed_errors = pNsc->MacStats.rxNoBufferCount;
	
	pAdpStats->tx_errors = pNsc->MacStats.txErrorCount;
	pAdpStats->tx_packets = pNsc->MacStats.txOkCount;
	pAdpStats->tx_dropped = pNsc->MacStats.txDroppedCount;
	pAdpStats->tx_carrier_errors = pNsc->MacStats.txCrsLostCount;
	pAdpStats->tx_aborted_errors = pNsc->MacStats.txAbortCount;
	pAdpStats->tx_fifo_errors = pNsc->MacStats.txUnderrunCount;
	pAdpStats->collisions = pNsc->MacStats.txTotCollisionCount;

	pAdpStats->rx_length_errors = 0;
	pAdpStats->rx_over_errors = 0;
	pAdpStats->rx_fifo_errors = 0;
	//pAdpStats->rx_compressed = 0;
	
	pAdpStats->tx_heartbeat_errors = 0;
	pAdpStats->tx_window_errors = 0;
	//pAdpStats->tx_compressed = 0;

	//pAdpStats->multicast = 0;

 	return pAdpStats;
}


/******************************************************************************
*f*
*   Name:           init_module 
*
*   Description:    init_module initializes the driver module. This routine 
*                   is called to initialize the driver when the module in 
*                   installed into a running kernel via insmod.
*
*   Parameters:     None 
*
*   Return Value:   OK or ENODEV 
*
*f*
***************************************************************************/

int
init_module (void)
{

#ifdef _DUMP
	struct module *mlist = &__this_module;
	printk("module %s is loaded at %p\n",mlist->name,
		&((char *)mlist)[mlist->size_of_struct]);
#endif

	return DrvProbe (NULL);
}

/******************************************************************************
*f*
*   Name:           cleanup_module 
*
*   Description:    cleanup_module initializes the driver module. This routine 
*                   is called to initialize the driver when the module in 
*                   installed into a running kernel via insmod.
*
*   Parameters:     None 
*
*   Return Value:   None 
*
*f*
***************************************************************************/

void
cleanup_module (void)
{
	struct device *cur_dev;
	struct device *pDev;

#ifdef NSCDEBUG
	printk(KERN_INFO "cleanup_module, called\n");
#endif

	/* Traverse through the list of device structures */
	cur_dev= DevList;
	while (cur_dev) 
	{
		/* Get the next device structure in the list */
		pDev = (struct device *) (((NsmContext *)cur_dev->priv)->next);

#ifdef NSCDEBUG
		printk(KERN_INFO "\nFreeing the IO region"); 
#endif

		/* Unregister the instance of the network device, corresponding to this 
			device structure  */
		unregister_netdev (cur_dev);
		
		/* Release the IO region held by this instance of the device */
		release_region(cur_dev->base_addr, PCI_IO_SIZE);

		kfree(((NsmContext *)cur_dev->priv)->pAdapterStats);
	
#ifdef NSCDEBUG	
		printk(KERN_INFO "Freeing the device\n"); 
#endif

		kfree(cur_dev); 
	
#ifdef NSCDEBUG	
		printk(KERN_INFO "Freed the device\n");
#endif
		
		/* Update the cur_dev to point to the next device in the list */
		cur_dev = pDev;
	}
}

/******************************************************************************
*f*
*   Name:           NsmTxCompleted 
*
*   Description:    This is called by HSM to indicate that transmission is 
*                   complete for a sublist of HsmTCBs in the HsmTCB list, 
*                   which is indicated by sending the first and the TCB next 
*                   to the last TCB in list.
*
*   Parameters:     pNsc:	Pointer to the AdapterContext structure
*
*                   TcbFirst:    Pointer to the first HsmTCB in the list for 
*                   which transmit completion has happened
*
*                   TcbLast:     Pointer to the HsmTCB which is the next from 
*                   the last TCB in the list for which transmit completion has 
*                   happened
* 
*
*   Return Value:   None 
*
*f*
***************************************************************************/

void NsmTxCompleted(AdapterContext *pNsc, HsmTCB *TcbFirst, HsmTCB *TcbLast)
{

	HsmTCB *tcb;

	tcb = TcbFirst;

	/* Free up the data buffer associated with the indicated HsmTCBs */  
	do
	{
		if((struct sk_buff *)tcb->pOsCtrlBlk) 
			kfree_skb((struct sk_buff *)tcb->pOsCtrlBlk);
		tcb->pOsCtrlBlk = NULL;
		tcb = tcb->pNext;
	}while(tcb != TcbLast);
}

/******************************************************************************
*f*
*   Name:           NsmRxPackets 
*
*   Description:    This is called by the HSM, to send the received packets in 
*                   a particular priority, over to the upper layers
*
*   Parameters:     pNsc:	Pointer to the AdapterContext structure
*
*                   RcbStart:	Pointer to the first HsmRCB in the list for 
*                   which receive completion has happened
*     
*                   RcbEnd:	Pointer to the HsmRCB which is the next from 
*                   the last RCB in the list for which receive completion has 
*                   happened
*
*                   priority:    Priority queue on which the packet was obtained
*
*   Return Value:   OK or ERROR 
*
*f*
***************************************************************************/

STATIC VOID 
NsmRxPackets (AdapterContext *pNsc, HsmRCB *RcbStart, HsmRCB *RcbEnd, 
	UCHAR Priority) 
{
	HsmRCB 			*rcb;
	HsmRCB 			*CurrRCB;
	HsmRCB 			*StartRCB;
	HsmRCB 			*TempRcb;
	struct sk_buff 	*NewSkb;
	struct sk_buff	*skb_new;
	struct sk_buff	*skb;
	UINT 			RemainingRCBs = 0;
	struct device 	*dev;
	UINT 			status=OK;
	UINT 			BufSizeReqd;
	static int 		NumPktsRcvd = 0;

	/* Get the pointer to the NsmContext */
	NsmContext *pNsm = (NsmContext *)pNsc->pNsmContext; 

#ifdef NSCDEBUG
	NsmDbgMsg("NsmRxPackets : Entered\n");
#endif

	/* Get the start of the list of HsmRCBs for which Rx complete has happened  
		*/
	rcb = RcbStart;

	/* Traverse through the HsmRCB list and for each RCB from RcbStart to 
		RcbEnd */
	while ( rcb != RcbEnd )
	{

#ifdef NSCDEBUG
		NsmDbgMsg("Processing the received packets\n");
#endif
	
		/* check for jumbo frames by checking the More flag in HsmRCB, if it's a
			jumbo packet */
		if(rcb->More)
		{

#ifdef NSCDEBUG  
			printk("Got a packet with the MORE bit set\n");
#endif
			
			/* Get the start of the HsmRCB list */
			StartRCB = rcb;

			/* Initialise the variable to update the size of the jumbo packet 
				received */
			BufSizeReqd = 0;

			/* For all the HsmRCBs into which this packet has spanned, do the 
				following */
			while( rcb->More) 
			{
				/* Calculate the size of the entire packet by adding the PktSize
					in the individual HsmRCBs, constituting the packet */
				BufSizeReqd += rcb->PktSize;

				/* Get the next HsmRCB in the RCB list */
				rcb = rcb->pNext;
				
				if(rcb == RcbEnd)
				{
#ifdef ASSERTION
					printk("The end of the packet not reached, bug!!!!!\n");
#endif
					return;
				}
			}
	
			CurrRCB = StartRCB;

			BufSizeReqd += rcb->PktSize;
			
			rcb = rcb->pNext;

			/* Allocate a new sk_buff to hold the complete jumbo packet, thereby
				one having a data buffer of enough size */
			NewSkb = alloc_skb(BufSizeReqd + HEAD_SPACE, GFP_ATOMIC);

			if( NewSkb == NULL)
			{

#ifdef ASSERTION
				printk("The allocation of the sk_buff of size %d for the jumbo packet FAILED\n", BufSizeReqd);
#endif
				/* Update statisitics */
				pNsc->MacStats.rxErrorCount++;
				pNsc->MacStats.rxDroppedCount++;

				continue;	
			}
			
			/* Set the priority field of the new sk_buff */
			NewSkb->priority = Priority;

#ifdef FTDEBUG
			printk("Processing the packet with the MORE bit set\n");
#endif

			/* Reserve some space at the beginning of the buffer */
			skb_reserve(NewSkb, HEAD_SPACE);

			dev = ( (struct sk_buff *) rcb->pOsCtrlBlk)->dev;

#ifdef FTDEBUG			
			printk("Reserved the space in the packet with the MORE bit set\n");
#endif

			/* Traverse through the HsmRCB list to copy the entire jumbo 
				packet into the newly allocated sk_buff */
			while(CurrRCB != rcb)
 			{
				memcpy( NewSkb->tail, 
					((struct sk_buff *)CurrRCB->pOsCtrlBlk)->data, 
						CurrRCB->PktSize);		
				skb_put(NewSkb, CurrRCB->PktSize); 

				CurrRCB->Status = RCB_FREE;
				
				CurrRCB = CurrRCB->pNext;
			}

			/* If checksum offloading is supported ie if it is already done in 
				hardware then indicate it in the skbuff */

#ifdef CHECKSUM
			NewSkb->ip_summed = CHECKSUM_UNNECESSARY; 
#endif
			NewSkb->dev = dev;

			NewSkb->protocol = eth_type_trans(NewSkb,dev);

#ifdef FTDEBUG	
			printk("Sending up the packet of length %d with the MORE bit set\n",NewSkb->len);
#endif

			/* Send the new sk_buff up */
			netif_rx(NewSkb);

#ifdef FTDEBUG	
			printk("Sent the packet up\n");
#endif
		}
		else
		{
			/* Through the HsmRCB, get the pointer to the OsCtrlBlk 
				( sk_buff ) */

#ifdef FTDEBUG
			NsmDbgMsg("NsmRxPackets : Got a normal packet\n");
#endif

			skb = (struct sk_buff *)rcb->pOsCtrlBlk;

			/* Clone this sk_buff */
			skb_new = skb_clone(skb, GFP_ATOMIC); 

			if( skb_new == NULL)
			{
				printk(KERN_ERR "NsmRxPackets : The cloning failed\n");
				TempRcb = rcb;

				/* Find out the number of packets remaining, which are now 
				going to be dropped*/
				while(TempRcb)
				{
					RemainingRCBs++;
					TempRcb = TempRcb->pNext;
				}	

				/* Update statistics */
				pNsc->MacStats.rxErrorCount += RemainingRCBs; 
				pNsc->MacStats.rxDroppedCount += RemainingRCBs; 

				status = ERROR;

				break;
			}
			
			rcb->Status = DUPED;

			skb_put(skb_new,rcb->PktSize);
			dev = skb_new->dev = skb->dev;
			
			skb_new->protocol = eth_type_trans(skb_new,dev);

#ifdef CHECKSUM
			skb_new->ip_summed = CHECKSUM_UNNECESSARY;			
#endif

			/* If checksum offloading is supported ie if it is already done in 
				hardware then indicate it in the skbuff */
			NumPktsRcvd++;
		
#ifdef FTDEBUG	
			printk("NsmRxPackets : Passing the packet no: %d of length %d up\n",
				NumPktsRcvd, skb_new->len); 
#endif

			netif_rx(skb_new);  

#ifdef FTDEBUG
			NsmDbgMsg("NsmRxPackets : Passed the packet up\n");
#endif

			/* Get the next HsmRCB in the RCB list */
			rcb = rcb->pNext;
		}
	}

	/* Set the time for the last receive */
	pNsm->dev->last_rx = jiffies;

}

/******************************************************************************
*f*
*   Name:           NsmMallocContiguous 
*
*   Description:    This is called by HSM to allocate a contiguous chunk of 
*                   memory, the virtual & physical addresses of the start of 
*                   the chunk of memory, is set in the Addrstruct structure. 
*                   The byte alignment required is given by the Alignment 
*                   parameter.
*
*   Parameters:     pAddrstruct: Pointer to the Addrstruct structure, where the 
*                   virtual and physical addresses need to be put
*
*                   Alignment:   The alignment factor, to which the memory 
*                   allocated needs to be aligned
*
*   Return Value:   SUCCESS or FAILURE 
*
*f*
***************************************************************************/

UCHAR  NsmMallocContiguous(VOID *pNsm,AddrStruct *pAddrStruct, 
	USHORT BitAlignment)
{

	VOID *pMem;

	UINT AlignedMem;

	USHORT Alignment = BitAlignment/8;
		
	pMem = kmalloc (pAddrStruct->BufSize + Alignment -1, 
		GFP_KERNEL | GFP_DMA);
	
	if(pMem == NULL)
		return FAILURE;
	
	NsmZeroMemory(pMem, pAddrStruct->BufSize + Alignment - 1);	

	AlignedMem = ((UINT) pMem) + (Alignment -1);

	pAddrStruct->pCookie = pMem;

	AlignedMem &= ~(Alignment -1);

	pAddrStruct->VirtAddr = (UCHAR  *)AlignedMem;

#ifdef ASSERTION
	if(AlignedMem & (Alignment-1))
		NsmDbgMsg1("MallocContiguous : The allocated memory is not %d byte aligned\n",Alignment);
#endif

	pAddrStruct->PhysAddr = (UCHAR *)virt_to_bus ((VOID *)AlignedMem);
	return SUCCESS;
	 
}

/******************************************************************************
*f*
*   Name:           NsmFreeContiguous
*
*   Description:    This is called by HSM to free a contiguous chunk of memory 
*
*   Parameters:     pAddrstruct: Pointer to the Addrstruct structure, where the 
*                   virtual and physical addresses are put
*
*   Return Value:   None
*
*f*
******************************************************************************/
			
void NsmFreeContiguous(VOID *pNsm, AddrStruct *pAddrStruct)
{
	kfree(pAddrStruct->pCookie);
	pAddrStruct->VirtAddr = 0;
	pAddrStruct->PhysAddr = 0;
	pAddrStruct->pCookie = 0;
}

/******************************************************************************
*f*
*   Name:           NsmAllocatePacket
*
*   Description:    This is called by HSM to allocate an sk_buff with a data 
*                   buffer associated with it.
*
*   Parameters:     pNsm:    Pointer to the NsmContext structure
*
*                   pRcb:    Pointer to the HsmRCB, to which the allocated 
*                   packet is to be linked to
*
*                   pAddrstruct: Pointer to the Addrstruct structure, where the 
*                   virtual and physical addresses are put
*
*   Return Value:   SUCCESS or FAILURE
*
*f*
******************************************************************************/

UCHAR  NsmAllocatePacket(VOID *pNsm, struct _HsmRCB *pRcb, AddrStruct *pAddr)
{

	struct sk_buff *skb;
	skb = alloc_skb(pAddr->BufSize, 
		GFP_KERNEL | GFP_DMA);
	
	if(skb == NULL)
	{
		printk("Allocation of the sk_buff failed\n");
		return FAILURE;
	}
	skb->dev = ((NsmContext *)pNsm)->dev;
	NsmZeroMemory(skb->data, pAddr->BufSize);	

	pAddr->pCookie = (VOID *)skb;
	pAddr->PhysAddr = (UCHAR *)virt_to_bus(skb->data);
	pRcb->pDataBuf = pAddr->VirtAddr = skb->data;

#ifdef ASSERTION
	if(((UINT)skb->data) & 7)
		NsmDbgMsg("AllocatePacket : The allocated memory is not 8 byte aligned\n");
#endif

	pRcb->pOsCtrlBlk = (VOID *)skb;
	return SUCCESS;
}

/******************************************************************************
*f*
*   Name:           NsmDeAllocatePacket
*
*   Description:    This is called by HSM to deallocate the sk_buff structure 
*                   and consequently the data buffer associated with it, within
*                   an HsmRCB
*
*   Parameters:     pRcb:    Pointer to the HsmRCB, to which the allocated 
*                   packet is to be linked to
*
*   Return Value:   None
*
*f*
******************************************************************************/

void NsmDeAllocatePacket(VOID *pNsm, AddrStruct *Addr, HsmRCB *pRcb)
{
	kfree_skb(pRcb->pOsCtrlBlk);
	pRcb->pOsCtrlBlk = NULL;
	pRcb->pDataBuf = NULL;
}

/* Utility routines */

/******************************************************************************
*f*
*   Name:         CreatePacketstruct
*
*   Description:  This creates a packet struct and encapsulates the data in 
*                 the sk_buff in to the Pktstruct
*
*   Parameters:   skb: Pointer to the sk_buff, from which the data to be sent 
                  down is taken to be encapsulated in to the Pktstruct data 
                  structure encapsulated in to the Pktstruct data structure
    
                  pPktSt:  Pointer to the Pktstruct structure, into which the 
                  data to be sent down is put
*
*   Return Value: None
*
*f*
******************************************************************************/


void 
CreatePacketStruct(struct sk_buff *skb, PktStruct *pPkt)
{
	UINT		Length;
	UCHAR 		*DataPtr;
	UINT		i=0;

	/*Get the length of the buffer that needs to be encapsulated within 
		the PktStruct */
	Length = skb->len;

	/* Get the pointer to the data in  the skbuff */
	DataPtr = skb->data;

	/* Put the data in each page as a new fragment */
	while(Length > PAGE_SIZE)
	{

#ifdef NSCDEBUG
		printk("Got a jumbo packet\n");
#endif
		/*Set the pointer to the physical memory for this fragment in the 
			packet structure */
		pPkt->PktArray[0].FragArray[i].pFrag = 
			(void *) virt_to_bus((void *)DataPtr);

		/* Update the data pointer */
		DataPtr += PAGE_SIZE;

		/* Update the length */
		Length -= PAGE_SIZE;

		pPkt->PktArray[0].FragArray[i].FragLen = PAGE_SIZE; 
		i++;
	}

	/* Set the pointer to the physical memory for this fragment in the packet 
		structure */
	pPkt->PktArray[0].FragArray[i].pFrag = 
		(UCHAR *)virt_to_bus(DataPtr);

	/* Update the length field of the fragment */
	pPkt->PktArray[0].FragArray[i].FragLen = Length; 

	i++;

	/* Put the number of fragments within the packet in the NumOfFrags field 
		within the packet struct */
	pPkt->PktArray[0].NumFrags = i;	

	/* Make the pOsCtrlBlk point to the skb */
	pPkt->PktArray[0].pOSCtrlBlk = skb;	
	pPkt->PktArray[0].ChkSum = 0;	

	pPkt->PktCount = 1;

}

#if 0 /* MKC */

VOID 
CreatePacketInfo(struct sk_buff *skb, PktInfo *pPkt)
{
	UINT	Length;
	UCHAR *DataPtr;
	UINT	i;

	/*Get the length of the buffer that needs to be encapsulated within 
		the PktStruct */
	Length = skb->len;

	/* Get the pointer to the data in  the skbuff */
	DataPtr = skb->data;

	/* Initialise the count for the number of fragments */
	i = 0;

	/* Put the data in each page as a new fragment */
	while(Length > PAGE_SIZE)
	{
		/*Set the pointer to the physical memory for this fragment in the 
			packet structure */
		pPkt->FragArray[i].pFrag = 
			(void *) virt_to_phys((void *)DataPtr);

		/* Update the data pointer */
		DataPtr += PAGE_SIZE;

		/* Update the length */
		Length -= PAGE_SIZE;

		pPkt->FragArray[i].FragLen = PAGE_SIZE; 
		i++;
	}

	/* Set the pointer to the physical memory for this fragment in the packet 
		structure */
	pPkt->FragArray[i].pFrag = 
		(UCHAR *)virt_to_phys(DataPtr);

	/* Update the length field of the fragment */
	pPkt->FragArray[i].FragLen = Length; 

	i++;

	/* Put the number of fragments within the packet in the NumOfFrags field 
		within the packet struct */
	pPkt->NumFrags = i;	

	/* Make the pOsCtrlBlk point to the skb */
	pPkt->pOSCtrlBlk = skb;	
	pPkt->ChkSum = 0;	

}
#endif

#if 1 /* MKC */
VOID 
CreatePacketInfo(struct sk_buff *skb, PktInfo *pPkt)
{

	/* Set the pointer to the physical memory for this fragment in the packet 
		structure */
	pPkt->FragArray[0].pFrag = 
		(UCHAR *)virt_to_phys(skb->data);

	/* Update the length field of the fragment */
	pPkt->FragArray[0].FragLen = skb->len; 

	/* Put the number of fragments within the packet in the NumOfFrags field 
		within the packet struct */
	pPkt->NumFrags = 1;	

	/* Make the pOsCtrlBlk point to the skb */
	pPkt->pOSCtrlBlk = skb;	
	pPkt->ChkSum = 0;	

}
#endif

/******************************************************************************
*f*
*   Name:         IsBrdCastOrMultiCast
*
*   Description:  This is used to check whether a particular address is a 
*                 broadcast or multicast address address
*
*   Parameters:   dev: Pointer to the device structure
*
*                 DestMac: Pointer to the destination MAC address
*
*   Return Value: SUCCESS or FAILURE
*
*f*
******************************************************************************/


UCHAR IsBrdCastOrMultiCast(struct device *dev, UCHAR *DestMac)
{
	struct dev_mc_list *mc_list;

#ifdef MKC /* This is not required */
	/* Check for broadcast address */
	if ((DestMac[0]==0xff) && (DestMac[1]==0xff) 
		&& (DestMac[2]==0xff) && (DestMac[3]==0xff) 
			&& (DestMac[4]==0xff) && (DestMac[5]==0xff))
		return SUCCESS;
#endif
	
	/* Check for multicast address */
	if(DestMac[0]==0x01)
	{

		mc_list = dev->mc_list;

		while(mc_list)
		{
			if(CMP_ETH_ADDR(mc_list->dmi_addr, (UCHAR *)DestMac, MAC_ADDR_LEN))
				return SUCCESS;
			
			mc_list = mc_list->next;
			
		}
		return FAILURE;
	}

	else
		return FAILURE;		    
}

/******************************************************************************
*f*
*   Name:         ChangeMtu
*
*   Description:  This changes the MTU size in the device structure to the new 
*                 value specified by NewMtu
*
*   Parameters:   dev:     Pointer to the device structure
*
*                 NewMtu:  The new MTU size which needs to be set in the device 
*                 structure
*
*   Return Value: EINVAL or OK
*
*f*
******************************************************************************/

STATIC int ChangeMtu(struct device *dev, int NewMtu)
{
	/* Check for the accepatable range of MTU sizes */
	if( (NewMtu < 68) || (NewMtu > 65280) )
		return -EINVAL;

	/* Set the mtu field in the device structure to the NewMtu if it is 
		falling within the acceptable range */
	dev->mtu = NewMtu;
	return (0);
}
		

inline UCHAR CMP_ETH_ADDR(UCHAR *SrcMac,UCHAR *DestMac, UINT Len)
{
	UINT i = 0;
	while(i<Len)
	{
       	if(DestMac[i]!=SrcMac[i])
			break;	
		i++;
	}
	if(i == Len)
		return 1;
    else
        return 0;
}                                                                              

inline VOID *NsmCreateLock()
{
	VOID *LockId = kmalloc(sizeof(spinlock_t), GFP_KERNEL);
	
	if(LockId == NULL)
		return NULL;
	
	spin_lock_init((spinlock_t *)LockId);	
	return LockId;
}

inline VOID NsmAcquireLock(AdapterContext *pAdapter,VOID *LockId)
{
	spin_lock((spinlock_t *)LockId);
}

inline VOID NsmReleaseLock(AdapterContext *pNsm, VOID *LockId)
{
	spin_unlock((spinlock_t *)LockId);
}

inline VOID *NsmMalloc(UINT size)
{
	VOID *ptr = kmalloc(size, GFP_KERNEL);
	
	if(ptr == NULL)
		return NULL;
	
	NsmZeroMemory(ptr,size);
	return ptr;
}
inline VOID PrintContents(UCHAR *start, UINT len)
{
	UCHAR	*ptr = start;
	UINT i;
	printk("The contents of the location :\n");
	for(i=0;i<len;i++)
	{
		printk(": %x :"	,*ptr);			
		ptr++;
	}
}
